;; Ebib v0.16
;;
;; Copyright (c) 2003, 2004, Joost Kremers
;; All rights reserved.
;;
;; Redistribution and use in source and binary forms, with or without
;; modification, are permitted provided that the following conditions
;; are met:
;;
;; 1. Redistributions of source code must retain the above copyright
;;    notice, this list of conditions and the following disclaimer.
;; 2. Redistributions in binary form must reproduce the above copyright
;;    notice, this list of conditions and the following disclaimer in the
;;    documentation and/or other materials provided with the distribution.
;; 3. The name of the author may not be used to endorse or promote products
;;    derived from this software without specific prior written permission.
;;
;; THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
;; IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
;; OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
;; IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
;; INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
;; NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES ; LOSS OF USE,
;; DATA, OR PROFITS ; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
;; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
;; (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
;; THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

(require 'cl)

(add-hook 'kill-emacs-query-functions 'ebib-kill-emacs-query-function)

;;;;;;;;;;;;;;;;;;;;;;
;; global variables ;;
;;;;;;;;;;;;;;;;;;;;;;

(defvar ebib-entry-types-hash nil "Holds the hash table containing the entry type definitions.")
(defvar ebib-ign-fields nil "Holds a list of the ignored fields.")
(defvar ebib-database nil "The main Ebib database.")
(defvar ebib-strings nil "Holds a hash table with the @STRING definitions.")
(defvar ebib-preamble nil "Holds a string the @PREAMBLE definition.")
(defvar ebib-number-of-entries nil "Number of entries stored in the database.")
;; note: ebib-keys-list is actually superfluous; the same info is stored in the keys buffer.
(defvar ebib-keys-list nil "Sorted list of the keys in the database.")
(defvar ebib-strings-list nil "Sorted list of the @STRING abbreviations.")
(defvar ebib-default-type nil "The default type for a newly created entry.")
(defvar ebib-file-name nil "Name of the BibTeX file.")
(defvar ebib-keys-buffer nil "The keys buffer.")
(defvar ebib-fields-buffer nil "The entry buffer.")
(defvar ebib-strings-buffer nil "The strings buffer.")
(defvar ebib-multiline-buffer nil "Buffer for editing multiline strings.")
(defvar ebib-help-buffer nil "Buffer showing Ebib help.")
(defvar ebib-current-entry nil "Points to the current entry in EBIB-KEYS-LIST.")
(defvar ebib-cur-entry-hash nil "The hash table containing the data of the current entry.
Only set when the user enters the entry buffer.")
(defvar ebib-cur-entry-fields nil "The fields of the type of the current
entry. Only set when the user enters the entry buffer.")
(defvar ebib-current-field nil "The current field.")
(defvar ebib-current-string nil "The current @STRING definition.")
(defvar ebib-minibuf-hist nil "Holds the minibuffer history for ebib")
(defvar ebib-saved-window-config nil "Stores the window configuration when Ebib is called.")
(defvar ebib-modified nil "T if the current database was modified.")
(defvar ebib-initialized nil "T if Ebib has been initialized.")
(defvar ebib-keys-overlay nil "Overlay to highlight the current entry.")
(defvar ebib-fields-overlay nil "Overlay to highlight the current field.")
(defvar ebib-strings-overlay nil "Overlay to highlight the current string.")
(defvar ebib-export-file-name nil "File name to export entries to.")
(defvar ebib-search-string nil "Stores the last search string.")
(defvar ebib-editing nil "Indicates what the user is editing.
Its value can be 'strings, 'fields, or 'preamble.")
(defvar ebib-multiline-raw nil "Indicates whether the multiline text being edited is raw.")
(defvar ebib-bibtex-identifier "[^\"#%'(),={}[:space:]]*" "Regex describing a licit BibTeX identifier.")
(defvar ebib-make-backup nil "Indicates whether a backup should be made of
the .bib file. This is only done when a newly loaded file is first saved.")
(defvar ebib-help "Stores the buffer the user was in when he displayed the help message.")

;;;;;;;;;;;;;;;;;;;;
;; things we need ;;
;;;;;;;;;;;;;;;;;;;;

;; note: what BibTeX calls required fields, i (accidentally) call obligatory fields...

;; macros allowing for easy definition of entry and field types.
(defmacro defentry (type obl-fields opt-fields)
  `(progn
     (puthash (quote ,type) (list (quote ,obl-fields) (quote ,opt-fields)) ebib-entry-types-hash)
     (if (not ebib-default-type) (setq ebib-default-type (quote ,type)))))

(defmacro def-ign-fields (&rest args)
  `(setq ebib-ign-fields (quote ,args)))

;; we sometimes need to walk through lists.
;; these macros yield the element directly preceding or following ELEM in
;; LIST. in order to work properly, ELEM must be unique in LIST,
;; obviously. if ELEM is the first/last element of LIST, the result is nil.
(defmacro next-elem (elem list)
  `(cadr (member ,elem ,list)))

(defmacro prev-elem (elem list)
  (let ((s-elem (gensym))
	(s-list (gensym)))
    `(let ((,s-elem ,elem)
	   (,s-list ,list))
       (if (equal ,s-elem (car ,s-list))
	   nil
	 (car (last ,s-list (1+ (length (member ,s-elem ,s-list)))))))))

(defmacro ebib-get-obl-fields (entry-type)
  `(car (gethash ,entry-type ebib-entry-types-hash)))

(defmacro ebib-get-opt-fields (entry-type)
  `(cadr (gethash ,entry-type ebib-entry-types-hash)))

;; this removes occurrences of \{ and \} from the input string.  note that
;; we cannot write "\{" as the delimiter for split-string, since that would
;; be read as "{". Similarly, "\\{" is impossible because \{ has a special
;; meaning in regexes. so this is the only work-around.
(defmacro remove-esc-braces (string)
  `(apply 'concat (split-string 
		   (apply 'concat (split-string ,string "[\\][{]"))
		   "[\\][}]")))

;; same, but for \".
(defmacro remove-esc-quotes (string)
  `(apply 'concat (split-string ,string "[\\][\"]")))

;; raw-p determines if a string is raw or not. the definition may seem a
;; bit unnatural, but alas it's really that difficult... first, one might
;; think a raw string is just a string that is not surrounded by {} or
;; "". but we cannot simply check if the first and last character of a
;; string are {...} or "...", because a string such as "{abc} # D # {efg}"
;; has { as first character and } as last, but is nonetheless
;; raw.

;; so the solution is this: remove every part from the string that is
;; between braces (or quotes), and see if anything's left. if there is, the
;; original string is raw. the removal is done with SPLIT-STRING. but
;; because strings surrounded by braces may contain embedded braces, we
;; must do this step repeatedly, in order to remove all levels of embedded
;; braces.

;; so what we do is the following: first we check if the string starts with
;; { or ". if not, it is certainly a raw string. (RAW-P detects this with
;; the default clause of COND.) second, if the string starts with {, we
;; first remove all \{ and \} from the string, because they would cause
;; SPLIT-STRING in the WHILE-clause to do the wrong thing. when that's
;; done, the WHILE-clause is executed. basically, what it does is take out
;; everything that matches the regex "{[^{]*?}". this regex translates to
;; the shortest sequence of characters starting with { and ending with },
;; and not containing another {. Essentially, this takes out the innermost
;; braces and their content. this procedure is then repeated until the
;; length of the list resulting from SPLIT-STRING is 1 or 0. in the first
;; case, the original string contained one or more sequences that were not
;; between braces, and hence the string is raw. in the latter case,
;; everything was between braces, and the string is not raw.

;; for example, take the string "{a{bc}d}". SPLIT-STRING takes out "{bc}",
;; returning the list ("{a" "d}"). because this list has more than one
;; element, the parts are concatenated again to form "{ad}", and WHILE
;; reiterates. In this next step, SPLIT-STRING will take out "{ad}",
;; because it is the smallest sequence of characters surrounded with {...}
;; not containing another {. the return value is then NIL.[1] because the
;; empty list has a length smaller than one, the WHILE loop will end with
;; SPLITTED set to NIL.

;; suppose now we have a string "{a{b}c} # D # {efg} # H". on the first
;; iteration, SPLIT-STRING will reduce this to ("{a" "c}" " # D # " " #
;; H"). this list has four elements, so the WHILE loop iterates again, now
;; on the string "{ac} # D # # H", the concatenation of the elements of the
;; list returned by SPLIT-STRING in the previous iteration. now,
;; SPLIT-STRING will return (" # D # # H"), which has only one element, so
;; WHILE will terminate.

;; for strings starting with " we do the same thing, except that we don't
;; have to iterate, because such strings cannot contain embedded strings
;; surrounded by quotes. (i.e., "ab"cd"ef" is not a licit BibTeX field
;; value, while {ab{cd}ef} is.

;; [1] annoying little fact: in emacs 21.3.1, this is the normal behaviour
;; of SPLIT-STRING. (i.e. (split-string "{a string}" "{[^{]*?}") ==> nil.)
;; in emacs 21.3.50.x, however, SPLIT-STRING needs an additional argument T
;; in order to obtain this behaviour. (the default behaviour being
;; (split-string "{a string}" "{[^{]*?}") ==> ("" "").)  thus making the
;; function incompatible between versions... <sigh>

(defun raw-p (string)
  "Non-nil if STRING is raw."
  (when (stringp string)
      (let (splitted
	    clear-str)
	(cond
	 ((eq (string-to-char string) ?\{)
	  (setq clear-str (remove-esc-braces string))
	  (while (progn
		   (setq splitted (delete "" (split-string clear-str "{[^{]*?}")))
		   (> (length splitted) 1))
	    (setq clear-str (apply 'concat splitted)))
	  (if splitted t nil))
	 ((eq (string-to-char string) ?\")
	  (setq clear-str (remove-esc-quotes string))
	  (setq splitted (delete "" (split-string clear-str "\".*?\"")))
	  (if splitted t nil))
	 (t t)))))

(defun to-raw (string)
  "Converts a string to its raw counterpart."
  (if (and (stringp string)
	   (not (raw-p string)))
      (substring string 1 -1)
    string))

(defun from-raw (string)
  "Converts a raw string to a non-raw one."
  (if (raw-p string)
      (concat "{" string "}")
    string))

(defun multiline-p (string)
  "True if STRING is multiline."
  (if (stringp string)
      (string-match "\n" string)))

(defun first-line (string)
  "Returns the first line of a multi-line string."
  (string-match "\n" string)
  (substring string 0 (match-beginning 0)))

(defmacro sort-in-buffer (limit str)
  "Finds the right position to insert STR in a buffer with lines sorted A-Z."
  (let ((lower (gensym))
	(middle (gensym))
	(upper (gensym))
	(beg (gensym))
	(s-limit (gensym)))
    `(let* ((,s-limit ,limit)
	    (,upper ,s-limit)
	    (,lower 0)
	    ,middle
	    ,beg)
       (when (> ,s-limit 0)
	 (setq ,upper ,s-limit)
	 (let ((,lower 0))
	   (goto-char (point-min))
	   (while (progn
		    (setq ,middle (/ (+ ,lower ,upper 1) 2))
		    (goto-line ,middle) ; if this turns out to be where we need to be,
		    (beginning-of-line) ; this puts the cursor at the right spot.
		    (> (- ,upper ,lower) 1)) ; if upper and lower differ by only 1, we have found the
					; position to insert the entry in.
	     (save-excursion
	       (let ((,beg (point)))
		 (end-of-line)
		 (if (string< (buffer-substring-no-properties ,beg (point)) ,str)
		     (setq ,lower ,middle)
		   (setq ,upper ,middle))))))))))

;; moving the overlays.
(defun ebib-set-keys-overlay ()
  (set-buffer ebib-keys-buffer)
  (beginning-of-line)
  (let ((beg (point)))
    (end-of-line)
    (move-overlay ebib-keys-overlay beg (point) ebib-keys-buffer)))

(defun ebib-set-fields-overlay ()
  (set-buffer ebib-fields-buffer)
  (beginning-of-line)
  (let ((beg (point)))
    (looking-at-goto-end "[^[:space:]]*")
    (move-overlay ebib-fields-overlay beg (point) ebib-fields-buffer)))

(defun ebib-set-strings-overlay ()
  (set-buffer ebib-strings-buffer)
  (beginning-of-line)
  (let ((beg (point)))
    (looking-at-goto-end "[^[:space:]]*")
    (move-overlay ebib-strings-overlay beg (point) ebib-strings-buffer)))

(defmacro with-buffer-writable (&rest body)
  "Makes the current buffer writable and executes the commands in BODY.
After BODY is executed, the buffer modified flag is unset."
  `(unwind-protect
       (let ((buffer-read-only nil))
	 ,@body)
     (set-buffer-modified-p nil)))

(defmacro ebib-redisplay-current-field ()
  "Redisplays the contents of the current field in the entry buffer."
  `(progn
     (set-buffer ebib-fields-buffer)
     (with-buffer-writable
      (goto-char (overlay-start ebib-fields-overlay))
      (let ((beg (point)))
	(end-of-line)
	(delete-region beg (point)))
      (insert (format "%-17s %s" (symbol-name ebib-current-field)
		      (ebib-get-field-highlighted ebib-current-field ebib-cur-entry-hash)))
      (ebib-set-fields-overlay))))

(defmacro ebib-redisplay-current-string ()
  "Redisplays the current string definition in the strings buffer."
    `(progn
       (set-buffer ebib-strings-buffer)
       (with-buffer-writable
	(let ((str (to-raw (gethash ebib-current-string ebib-strings))))
	  (goto-char (overlay-start ebib-strings-overlay))
	  (let ((beg (point)))
	    (end-of-line)
	    (delete-region beg (point)))
	  (insert (format "%-18s %s" ebib-current-string 
			  (if (multiline-p str)
			      (concat "+" (first-line str))
			    (concat " " str))))
	  (ebib-set-strings-overlay)))))

(defmacro symbol-or-string (x)
  "Returns the symbol-name of X if X is a symbol, otherwise return X.

Much like SYMBOL-NAME, except it does not throw an error if X is not a
symbol."
  `(if (symbolp ,x)
      (symbol-name ,x)
    ,x))

(defun looking-at-goto-end (str &optional match)
  "Like LOOKING-AT but moves point to the end of the matching string.

MATCH acts just like the argument to MATCH-END, and defaults to 0."
  (or match (setq match 0))
  (let ((case-fold-search t))
    (if (looking-at str)
	(goto-char (match-end match)))))

(defun ebib-create-collection (hashtable)
  "Creates a list from the keys in HASHTABLE that can be used as COLLECTION in COMPLETING-READ.

The keys of HASHTABLE must be either symbols or strings."
  (let ((result nil))
    (maphash '(lambda (x y)
		(setq result (cons (cons (symbol-or-string x)
					 0)
				   result)))
	     hashtable)
    result))

(defmacro match-all (match-str string)
  "Highlights all the matches of match-str in string.

The return value is a list of two elements: the first is the modified
string, the second either t or nil, indicating whether a match was found at
all."
  (let ((counter (gensym))
	(s-string (gensym))
	(s-match-str (gensym)))
    `(let ((,s-string ,string)
	   (,s-match-str ,match-str))
       (do ((,counter 0 (match-end 0)))
	   ((not (string-match ,s-match-str ,s-string ,counter)) (values ,s-string (not (= ,counter 0))))
	 (add-text-properties (match-beginning 0) (match-end 0) '(face highlight) ,s-string)))))

(defun ebib-get-field-highlighted (field current-entry &optional match-str)
  ;; note: we need to work on a copy of the string, otherwise the highlights
  ;; are made to the string as stored in the database. hence copy-sequence.
  (let ((case-fold-search t)
	(string (copy-sequence (gethash field current-entry)))
	(raw " ")
	(multiline " ")
	(matched nil))
    ;; we have to do a couple of things now:
    ;; - remove {} or "" around the string, if they're there
    ;; - search for match-str
    ;; - properly adjust the string if it's multiline
    ;; but all this is not necessary if there was no string
    (if (null string)
	(setq string "")
      (if (raw-p string)
	  (setq raw "*")
	(setq string (to-raw string)))   ; we have to make the string look nice
      (when match-str
	(multiple-value-setq (string matched) (match-all match-str string)))
      (when (multiline-p string)
	;; IIUC propertize shouldn't be necessary here, as the variable
	;; multiline is local and therefore the object it refers to should
	;; be GC'ed when the function returns. but for some reason, the
	;; plus sign is persistent, and if it's been highlighted as the
	;; result of a search, it stays that way.
	(setq multiline (propertize "+" nil nil))
	(setq string (first-line string))))
    (when (and matched
	       (string= multiline "+"))
      (add-text-properties 0 1 '(face highlight) multiline))
    (concat raw multiline string)))

(defun ebib-fill-fields-buffer (&optional match-str)
  "Fills the entry buffer with the fields of the current entry.

MATCH-STRING is a regexp that will be highlighted when it occurs in the
field contents."
  (set-buffer ebib-fields-buffer)
  (with-buffer-writable
    (erase-buffer)
    (when ebib-current-entry
      (let* ((current-entry (gethash (car ebib-current-entry) ebib-database))
	     (entry-type (gethash 'type* current-entry))
	     (obl-fields (ebib-get-obl-fields entry-type))
	     (opt-fields (ebib-get-opt-fields entry-type)))
	(insert (format "%-19s %s\n" "type" entry-type))
	(mapcar '(lambda (fields)
		   (insert "\n")
		   (mapcar '(lambda (x)
			      (insert
			       (format "%-17s %s\n" x (or
						       (ebib-get-field-highlighted x current-entry match-str)
						       ""))))
			   fields))
		(list obl-fields opt-fields ebib-ign-fields)))
      (setq ebib-current-field 'type*)
      (goto-char (point-min))
      (ebib-set-fields-overlay))))

(defun ebib-set-modified (mod)
  "Sets the modified flag of the keys buffer and the flag ebib-modified to
MOD. MOD must be either T or NIL."
  (save-excursion
    (set-buffer ebib-keys-buffer)
    (set-buffer-modified-p mod)
    (setq ebib-modified mod)))

;; this is ugly. i need to disable C-x k and C-x b in the fields and
;; strings buffers. unsetting them locally has no effect, so i locally set
;; them to ebib-nil...

(defun ebib-t ()
  "Do nothing."
  (interactive)
  t)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; main program execution ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun ebib ()
  "Ebib, a BibTeX database manager."
  (interactive)
  (if (or (equal (window-buffer) ebib-keys-buffer)
	  (equal (window-buffer) ebib-fields-buffer))
      (error "Ebib already active")
    (unless ebib-initialized
      (ebib-init))
    ;; we save the current window configuration.
    (setq ebib-saved-window-config (current-window-configuration))
    ;; create the window configuration we want for ebib.
    (delete-other-windows)
    (switch-to-buffer ebib-keys-buffer)
    (let* ((keys-window (selected-window))
	   (entry-window (split-window keys-window 11)))
      (set-window-buffer entry-window ebib-fields-buffer))))

(defun ebib-create-buffers ()
  "Creates the buffers for Ebib."
  ;; first we create a buffer for multiline editing.  this one does *not*
  ;; have a name beginning with a space, because undo-info is normally
  ;; present in an edit buffer.
  (setq ebib-multiline-buffer (get-buffer-create "*Ebib-edit*"))
  (set-buffer ebib-multiline-buffer)
  (ebib-multiline-edit-mode)
  ;; then we create a buffer to hold the fields of the current entry.
  (setq ebib-fields-buffer (get-buffer-create " *Ebib-entry*"))
  (set-buffer ebib-fields-buffer)
  (ebib-fields-mode)
  ;; then we create a buffer to hold the @STRING definitions
  (setq ebib-strings-buffer (get-buffer-create " *Ebib-strings*"))
  (set-buffer ebib-strings-buffer)
  (ebib-strings-mode)
  ;; then we create the help buffer
  (setq ebib-help-buffer (get-buffer-create " *Ebib-help*"))
  (set-buffer ebib-help-buffer)
  (ebib-help-mode)
  ;; and lastly we create a buffer for the entry keys.
  (setq ebib-keys-buffer (get-buffer-create " *Ebib*"))
  (set-buffer ebib-keys-buffer)
  (ebib-keys-mode))

(defun ebib-init ()
  "Initialises Ebib.

This function sets all variables to their initial values, creates the
buffers and reads the rc file."
  (setq ebib-entry-types-hash (make-hash-table)
	ebib-ign-fields nil
	ebib-database (make-hash-table :test 'equal)
	ebib-strings (make-hash-table :test 'equal)
	ebib-number-of-entries 0
	ebib-keys-list nil
	ebib-file-name nil
	ebib-make-backup nil
	ebib-current-entry nil
	ebib-cur-entry-hash nil
	ebib-current-field nil
	ebib-minibuf-hist nil
	ebib-saved-window-config nil
	ebib-modified nil)
  (load "~/.ebibrc")
  (ebib-create-buffers)
  (setq ebib-keys-overlay (make-overlay 1 1 ebib-keys-buffer))
  (overlay-put ebib-keys-overlay 'face 'highlight)
  (setq ebib-fields-overlay (make-overlay 1 1 ebib-fields-buffer))
  (overlay-put ebib-fields-overlay 'face 'highlight)
  (setq ebib-strings-overlay (make-overlay 1 1 ebib-strings-buffer))
  (overlay-put ebib-strings-overlay 'face 'highlight)
  (setq ebib-initialized t))

(defun ebib-quit ()
  "Quits Ebib.
The Ebib buffers are killed, all variables except the keymaps are set to nil."
  (interactive)
  (when (if ebib-modified
	    (yes-or-no-p "Database modified. Quit anyway? ")
	  (y-or-n-p "Quit Ebib? "))
    (kill-buffer ebib-fields-buffer)
    (kill-buffer ebib-keys-buffer)
    (kill-buffer ebib-multiline-buffer)
    (setq ebib-entry-types-hash nil
	  ebib-ign-fields nil
	  ebib-database nil
	  ebib-strings nil
	  ebib-preamble nil
	  ebib-number-of-entries nil
	  ebib-default-type nil
	  ebib-keys-list nil
	  ebib-strings-list nil
	  ebib-file-name nil
	  ebib-make-backup nil
	  ebib-keys-buffer nil
	  ebib-fields-buffer nil
	  ebib-current-entry nil
	  ebib-modified nil
	  ebib-initialized nil
	  ebib-keys-overlay nil
	  ebib-fields-overlay nil
	  ebib-export-file-name nil)
    (set-window-configuration ebib-saved-window-config)
    (message "")))

(defun ebib-kill-emacs-query-function ()
  "Ask if the user wants to save the database loaded in Ebib when Emacs is
killed and the database has been modified."
  (if (not ebib-modified)
      t
    (if (y-or-n-p "Save Ebib database? ")
	(progn
	  (ebib-save-database)
	  t)
      (yes-or-no-p "Ebib database was modified. Kill anyway? "))))

;; the next two functions are used in loading the database. a database is
;; loaded by reading the raw file into a temp buffer and then reading all
;; the entries in it. for this we need to be able to search a matching
;; parenthesis and a matching double quote.

(defun ebib-match-paren-forward (limit)
  "Moves forward to the closing parenthesis matching the opening parenthesis at POINT.

Does not search/move beyond LIMIT. Returns T if a matching parenthesis was
found, NIL otherwise. If point was not at an opening parenthesis at all,
NIL is returned and point is not moved. If point was at an opening
parenthesis but no matching closing parenthesis was found, point is moved
to LIMIT."
  (if (not (or (eq (char-after) ?\{)
	       (eq (char-after) ?\")))
      nil
    (save-restriction
      (narrow-to-region (point-min) limit)
      (condition-case nil
	  (progn
	    (forward-list)
	    ;; all of ebib expects that point moves to the closing
	    ;; parenthesis, not right after it, so we adjust.
	    (forward-char -1)
	    t) ; return t because a matching brace was found
	(error (progn
		 (goto-char (point-max)) ; point-max because the narrowing is still in effect
		 nil))))))

(defun ebib-match-quote-forward (limit)
  "Moves to the closing double quote matching the quote at POINT.

Does not search/move beyond LIMIT. Returns T if a matching quote was found,
NIL otherwise. If point was not at a double quote at all, NIL is returned
and point is not moved. If point was at a quote but no matching closing
quote was found, point is moved to LIMIT."
  (when (eq (char-after (point)) ?\") ; make sure we're on a double quote.
    (while
	(progn
	  (forward-char) ; we need to move forward because we're on a double quote.
	  (skip-chars-forward "^\"" limit) ; find the next double quote.
	  (and (eq (char-before) ?\\)  ; if it's preceded by a backslash,
	       (< (point) limit)))) ; and we're still below LIMIT, keep on searching.
    (eq (char-after (point)) ?\"))) ; return T or NIL based on whether we've found a quote.

;;;;;;;;;;;;;;;
;; keys-mode ;;
;;;;;;;;;;;;;;;

(defvar ebib-keys-mode-map
  (make-keymap) "Keymap for the ebib keys buffer.")
(suppress-keymap ebib-keys-mode-map)

(define-key ebib-keys-mode-map [up] 'ebib-prev-entry)
(define-key ebib-keys-mode-map [down] 'ebib-next-entry)
(define-key ebib-keys-mode-map "k" 'ebib-prev-entry)
(define-key ebib-keys-mode-map "j" 'ebib-next-entry)
(define-key ebib-keys-mode-map "z" 'ebib-lower)
(define-key ebib-keys-mode-map "e" 'ebib-edit-entry)
(define-key ebib-keys-mode-map "g" 'ebib-goto-first-entry)
(define-key ebib-keys-mode-map "G" 'ebib-goto-last-entry)
(define-key ebib-keys-mode-map [home] 'ebib-goto-first-entry)
(define-key ebib-keys-mode-map [end] 'ebib-goto-last-entry)
(define-key ebib-keys-mode-map "b" 'ebib-entry-page-up)
(define-key ebib-keys-mode-map " " 'ebib-entry-page-down)
(define-key ebib-keys-mode-map [prior] 'ebib-entry-page-up)
(define-key ebib-keys-mode-map [next] 'ebib-entry-page-down)
(define-key ebib-keys-mode-map "q" 'ebib-quit)
(define-key ebib-keys-mode-map "s" 'ebib-save-database)
(define-key ebib-keys-mode-map "w" 'ebib-write-database)
(define-key ebib-keys-mode-map "c" 'ebib-close-database)
(define-key ebib-keys-mode-map "o" 'ebib-load-bibtex-file)
(define-key ebib-keys-mode-map "m" 'ebib-merge-bibtex-file)
(define-key ebib-keys-mode-map "a" 'ebib-add-entry)
(define-key ebib-keys-mode-map "/" 'ebib-search)
(define-key ebib-keys-mode-map "n" 'ebib-search-next)
(define-key ebib-keys-mode-map "f" 'ebib-print-file-name)
(define-key ebib-keys-mode-map "d" 'ebib-delete-entry)
(define-key ebib-keys-mode-map [return] 'ebib-select-entry)
(define-key ebib-keys-mode-map "x" 'ebib-export-entry)
(define-key ebib-keys-mode-map "S" 'ebib-edit-strings)
(define-key ebib-keys-mode-map "P" 'ebib-edit-preamble)
(define-key ebib-keys-mode-map "X" 'ebib-export-preamble)
(define-key ebib-keys-mode-map "h" 'ebib-keys-help)

(define-derived-mode ebib-keys-mode
  fundamental-mode "Ebib-keys"
  "Major mode for the Ebib keys buffer."
  ;; we redefine these two keys because an ebib buffer shouldn't be killed
  ;; or switched from. it's probably cheating to do it this way instead of
  ;; with define-key, but what the heck. ;-)
  (setq buffer-read-only t)
  (local-set-key "\C-xb" 'ebib-lower)
  (local-set-key "\C-xk" 'ebib-quit))

(defun ebib-fill-keys-buffer ()
  "Fills the keys buffer with the list of keys."
  (set-buffer ebib-keys-buffer)
  (with-buffer-writable
   (erase-buffer)
   (when ebib-current-entry
     (mapcar '(lambda (x) (insert (format "%s\n" x))) ebib-keys-list)
     (goto-char (point-min))
     (re-search-forward (format "^%s$" (car ebib-current-entry))))
   (beginning-of-line)
   (ebib-set-keys-overlay)
   (set-buffer-modified-p nil)))

(defun ebib-load-bibtex-file ()
  "Loads a BibTeX file into ebib."
  (interactive)
  (if (or ebib-file-name ebib-modified)
      (error "Database not empty. Use `m' to merge a file ")
    (let ((file (read-file-name "File to load: ")))
      ;; we load the file by putting it in a buffer and then parsing it.
      (with-temp-buffer
	(let ((full-file-name (insert-file-contents file)))
	  ;; if we don't find any entries, issue a warning
	  (if (= (setq ebib-number-of-entries (ebib-find-bibtex-entries)) 0)
	      (error "No BibTeX entries found in file")
	    ;; otherwise set some variables
	    (setq ebib-file-name (car full-file-name))
	    (setq ebib-make-backup t)
	    (setq ebib-keys-list (sort ebib-keys-list 'string<))
	    (when ebib-strings-list
	      (setq ebib-strings-list (sort ebib-strings-list 'string<)))
	    (setq ebib-current-entry ebib-keys-list)
	    ;; and fill the buffers. note that filling a buffer also makes
	    ;; that buffer active. therefore we put ebib-fill-keys-buffer
	    ;; last.
	    (ebib-fill-fields-buffer)
	    (ebib-fill-keys-buffer)))))))

(defun ebib-merge-bibtex-file ()
  "Merges a BibTeX file into the database."
  ;; this does more or less the same as ebib-load-bibtex-file, but some
  ;; details are different.
  (interactive)
    (let ((file (read-file-name "File to merge: ")))
      (with-temp-buffer
	(let ((full-file-name (insert-file-contents file)))
	  (if (= (ebib-find-bibtex-entries) 0)
	      (error "No BibTeX entries found in file")
	    (setq ebib-keys-list (sort ebib-keys-list 'string<))
	    (setq ebib-number-of-entries (length ebib-keys-list))
	    (when ebib-strings-list
	      (setq ebib-strings-list (sort ebib-strings-list 'string<)))
	    (setq ebib-current-entry ebib-keys-list)
	    (ebib-fill-fields-buffer)
	    (ebib-fill-keys-buffer)
	    (ebib-set-modified t))))))

(defun ebib-find-bibtex-entries ()
  "Finds the BibTeX entries in the current buffer.

The search is started at the beginnig of the buffer. All entries found are
stored in the hash table ebib-database. Returns the number of entries
found, or nil if no entries were found."
  (let ((n-entries 0))
    (goto-char (point-min))
    (while (re-search-forward "^@" nil t)     ; find the next entry
      (let ((beg (point))
	    (entry-type)
	    (entry-key)
	    (entry-limit))
	(when (looking-at-goto-end (concat ebib-bibtex-identifier "[\(\{]"))
	  (setq entry-type (downcase (buffer-substring-no-properties beg (1- (point)))))
	  (cond 
	   ((equal entry-type "string") (ebib-read-string))          ; string and preamble must be treated differently
	   ((equal entry-type "preamble") (ebib-read-preamble))
	   ((equal entry-type "comment") (ebib-match-paren-forward (point-max)))  ; ignore comments
	   ((gethash (intern-soft entry-type) ebib-entry-types-hash)       ; if the entry type has been defined
	    (setq beg (point))
	    (setq entry-limit (save-excursion
				(backward-char)
				(ebib-match-paren-forward (point-max))
				(point)))
	    ;; assuming that abbreviations can be composed of letters,
	    ;; numbers, colon and hyphen, and must begin with a letter.  we
	    ;; should find out if this is really the case. same for the
	    ;; entry type, field type and string abbreviations!
	    (when (looking-at-goto-end (concat ebib-bibtex-identifier "[[:space:]]*,")) ; this delimits the entry key
		(setq entry-key (buffer-substring-no-properties beg (1- (point))))
		(if (member entry-key ebib-keys-list)
		    (message (format "Entry `%s' duplicated " entry-key))
		  (let ((fields (ebib-find-bibtex-fields (intern-soft entry-type) entry-limit)))
		    ;; if fields were found, we store them.
		    (when fields
		      (puthash entry-key fields ebib-database)
		      (setq ebib-keys-list (cons entry-key ebib-keys-list))
		      (setq n-entries (1+ n-entries)))))))
	   ;; this is the catch-all of the COND above
	   (t (message (format "Unknown entry type `%s'. Skipping." entry-type)) ; we found something we don't know
	       (ebib-match-paren-forward (point-max)))))))
    n-entries))

(defun ebib-read-string ()
  "Reads the @string definition beginning at the line POINT is on.

If a proper abbreviation and string are found, they are stored in the
database. Returns the string if one was read, nil otherwise."
  (let ((limit (save-excursion  ; we find the matching end parenthesis
		 (backward-char)
		 (ebib-match-paren-forward (point-max))
		 (point))))
    (skip-chars-forward "\"#%'(),={} \n\t\f" limit)  ; "^a-zA-Z"
    (let ((beg (point))
	  abbr
	  string)
      (when (looking-at-goto-end (concat "\\(" ebib-bibtex-identifier "\\)[[:space:]]*=") 1)
	(setq abbr (buffer-substring-no-properties beg (point)))
	(when (equal abbr "")
	  (setq abbr nil))
	(skip-chars-forward "^\"{" limit)
	(let ((beg (point)))
	  (setq string (cond
			((eq (char-after) ?\" ) (if (ebib-match-quote-forward limit)
						    (buffer-substring-no-properties beg (1+ (point)))
						  nil))
			((eq (char-after) ?\{ ) (if (ebib-match-paren-forward limit)
						    (buffer-substring-no-properties beg (1+ (point)))
						  nil))
			(t nil)))
	  (when (and abbr string)
	    (if (member abbr ebib-strings-list)
		(message (format "@STRING definition `%s' duplicated" abbr))
	      (puthash abbr string ebib-strings)
	      (setq ebib-strings-list (cons abbr ebib-strings-list)))))))))

(defun ebib-read-preamble ()
  "Reads the @preamble definition.

If there was already another @preamble definition, the new one is added to
the existing one with a hash sign # between them."
  (let ((beg (point)))
    (forward-char -1)
    (ebib-match-paren-forward (point-max))
    (let ((text (buffer-substring-no-properties beg (point))))
      (if ebib-preamble
	  (setq ebib-preamble (concat ebib-preamble "\n# " text))
	(setq ebib-preamble text)))))

(defun ebib-find-bibtex-fields (entry-type limit)
  "Finds the fields of the BibTeX entry that starts on the line POINT is on.

Returns a hash table containing all the fields and values, or NIL if none
were found. ENTRY-TYPE is the type of the entry, which will be recorded in
the hash table. Before the search starts, POINT is moved back to the
beginning of the line."
  (beginning-of-line)
  (let ((fields (make-hash-table))
  	(fields-found nil))
    (while (progn
	     (skip-chars-forward "^," limit) ; we must move to the next comma,
	     (eq (char-after) ?,))         ; and make sure we are really on a comma.
      (skip-chars-forward "\"#%'(),={} \n\t\f" limit)  ; "^a-zA-Z"
      (let ((beg (point))
	    (field-type))
	(when (looking-at-goto-end (concat "\\(" ebib-bibtex-identifier "\\)[[:space:]]*=") 1)
	  (setq field-type (intern (downcase (buffer-substring-no-properties beg (point)))))
	  (unless (eq field-type 'type*) ; the 'type*' key holds the entry type, so we can't use it;
					 ; i named it 'type*' so people can still use 'type' as an
					 ; field type, if they so like
	    (let ((field-contents (ebib-get-field-contents limit)))
	      (when field-contents
		(puthash field-type field-contents fields)
		(setq fields-found t)))))))
    (when fields-found
      (puthash 'type* entry-type fields)
      fields)))

(defun ebib-get-field-contents (limit)
  "Gets the contents of a BibTeX field.

LIMIT indicates the end of the entry, beyond which the function will not 
search."
  (skip-chars-forward "#%'(),=} \n\t\f" limit)  ; "^{\"0-9a-zA-Z"
  (let ((beg (point)))
    (buffer-substring-no-properties beg (ebib-find-end-of-field limit))))

(defun ebib-find-end-of-field (limit)
  "Moves POINT to the end of a field's contents and returns POINT.

The contents of a field is delimited by a comma or by the closing brace of
the entry. The latter is at position LIMIT."
  (while (and (not (eq (char-after) ?\,))
	      (< (point) (1- limit)))
    (cond
     ((eq (char-after) ?\{) (ebib-match-paren-forward limit))
     ((eq (char-after) ?\") (ebib-match-quote-forward limit)))
    (forward-char 1))
  (point))

(defun ebib-lower ()
  "Hides the Ebib buffers, but does not delete them."
  (interactive)
  (if (not (or (equal (window-buffer) ebib-keys-buffer)
	       (equal (window-buffer) ebib-fields-buffer)))
      (error "Ebib is not active")
    (set-window-configuration ebib-saved-window-config)
    (bury-buffer ebib-fields-buffer)
    (bury-buffer ebib-keys-buffer)
    (bury-buffer ebib-multiline-buffer)
    (bury-buffer ebib-strings-buffer)))

(defun ebib-prev-entry ()
  "Moves to the previous BibTeX entry."
  (interactive)
  (if (eq ebib-current-entry ebib-keys-list) ; if the current entry is the first entry,
      (beep)                                 ; just beep.
    (setq ebib-current-entry (last ebib-keys-list (1+ (length ebib-current-entry))))
    (goto-char (overlay-start ebib-keys-overlay))
    (forward-line -1)
    (ebib-set-keys-overlay)
    (ebib-fill-fields-buffer)))

(defun ebib-next-entry ()
  "Moves to the next BibTeX entry."
  (interactive)
  (if (= (length ebib-current-entry) 1) ; if we're on the last entry,
      (beep)                            ; just beep.
    (setq ebib-current-entry (last ebib-keys-list (1- (length ebib-current-entry))))
    (goto-char (overlay-start ebib-keys-overlay))
    (forward-line 1)
    (ebib-set-keys-overlay)
    (ebib-fill-fields-buffer)))

(defun ebib-add-entry (entry-key)
  "Adds a new entry to the database."
  (interactive "sNew entry key: ")
  ;; to add a new entry, we insert it not only in the database and
  ;; ebib-keys-list, but also in the keys buffer. that is, we do not erase
  ;; the keys buffer to refill it after the entry has been added. for this
  ;; reason, we need to find the right spot in the buffer to insert the
  ;; entry.
  (set-buffer ebib-keys-buffer)
  (if (equal entry-key "")
      nil
    (if (member entry-key ebib-keys-list)
	(error "Key already exists")
      (sort-in-buffer (1+ ebib-number-of-entries) entry-key)
      (with-buffer-writable
       (insert (format "%s\n" entry-key)))  ; add the entry in the buffer.
      (forward-line -1) ; move one line up to position the cursor on the new entry.
      (ebib-set-keys-overlay)
      (puthash entry-key (make-hash-table) ebib-database)
      (puthash 'type* ebib-default-type (gethash entry-key ebib-database)) ; give it a default type
      (setq ebib-keys-list (sort (cons entry-key ebib-keys-list) 'string<))
      (setq ebib-number-of-entries (1+ ebib-number-of-entries))
      (setq ebib-current-entry (member entry-key ebib-keys-list))
      ;; fill the entry buffer and switch to it.
      (ebib-fill-fields-buffer)
      (ebib-edit-entry)
      (ebib-set-modified t))))

(defun ebib-close-database ()
  "Closes the BibTeX database loaded in Ebib."
  (interactive)
  (when (or ebib-modified ebib-file-name)
    (when (if ebib-modified
	      (yes-or-no-p "Database modified. Close it anyway? ")
	    (y-or-n-p "Close database? "))
      (clrhash ebib-database)
      (clrhash ebib-strings)
      (setq ebib-preamble nil)
      (setq ebib-number-of-entries 0
	    ebib-file-name nil
	    ebib-make-backup nil
	    ebib-keys-list nil
	    ebib-strings-list nil
	    ebib-current-entry nil)
      ;; fields buffer
      (set-buffer ebib-fields-buffer)
      (with-buffer-writable
       (erase-buffer))
      (delete-overlay ebib-keys-overlay)
      ;; keys buffer
      (set-buffer ebib-keys-buffer)
      (with-buffer-writable
       (erase-buffer))
      (delete-overlay ebib-keys-overlay)
      ;; strings buffer
      (set-buffer ebib-strings-buffer)
      (with-buffer-writable
       (erase-buffer))
      (delete-overlay ebib-strings-overlay)
      ;; multiline edit buffer
      (set-buffer ebib-multiline-buffer)
      (with-buffer-writable
       (erase-buffer))
      (ebib-set-modified nil)
      (message "Database closed."))))

(defun ebib-goto-first-entry ()
  "Moves to the first BibTeX entry in the database."
  (interactive)
  (setq ebib-current-entry ebib-keys-list)
  (set-buffer ebib-keys-buffer)
  (goto-char (point-min))
  (ebib-set-keys-overlay)
  (ebib-fill-fields-buffer))

(defun ebib-goto-last-entry ()
  "Moves to the last entry in the BibTeX database."
  (interactive)
  (setq ebib-current-entry (last ebib-keys-list))
  (set-buffer ebib-keys-buffer)
  (goto-line ebib-number-of-entries)
  (ebib-set-keys-overlay)
  (ebib-fill-fields-buffer))

(defun ebib-edit-entry ()
  "Edits the current BibTeX entry."
  (interactive)
  (when ebib-current-entry
    (setq ebib-cur-entry-hash (gethash (car ebib-current-entry) ebib-database))
    (setq ebib-cur-entry-fields
	  (let* ((entry-type (gethash 'type* ebib-cur-entry-hash))
		 (fields (cons 'type* (append (ebib-get-obl-fields entry-type)
					      (ebib-get-opt-fields entry-type) 
					      ebib-ign-fields))))
	    fields))
    (setq ebib-editing 'fields)
    (other-window 1)
    (switch-to-buffer ebib-fields-buffer)
    (goto-char (overlay-end ebib-fields-overlay))))

(defun ebib-entry-page-up ()
  "Moves 10 entries up in the database."
  (interactive)
  (if (<= (- ebib-number-of-entries (length ebib-current-entry)) 10)
      (ebib-goto-first-entry)
    (setq ebib-current-entry (nthcdr
			      (- ebib-number-of-entries (length ebib-current-entry) 10)
			      ebib-keys-list))
    (goto-char (overlay-start ebib-keys-overlay))
    (forward-line -10)
    (ebib-set-keys-overlay)
    (ebib-fill-fields-buffer)))

(defun ebib-entry-page-down ()
  "Moves 10 entries down in the database."
  (interactive)
  (if (<= (length ebib-current-entry) 10)
      (ebib-goto-last-entry)
    (setq ebib-current-entry (nthcdr
			      (- ebib-number-of-entries (length ebib-current-entry) -10)
			      ebib-keys-list))
    (goto-char (overlay-start ebib-keys-overlay))
    (forward-line 10)
    (ebib-set-keys-overlay)
    (ebib-fill-fields-buffer)))

(defun ebib-format-entry (key)
  "Formats an entry in the Ebib database in BibTeX format."
  (let ((entry (gethash key ebib-database)))
    (when entry
      (insert (format "@%s{%s,\n" (gethash 'type* entry) key))
      (maphash '(lambda (key value)
		  (unless (eq key 'type*)
		    (insert (format "\t%s = %s,\n" key value))))
	       entry)
      (delete-char -2)  ; the final ",\n" must be deleted
      (insert "\n}\n\n"))))

(defun ebib-format-strings ()
  "Formats the @string commands in the database."
  (maphash '(lambda (key value)
	      (insert (format "@string{%s = %s}\n" key value)))
	   ebib-strings)
  (insert "\n"))

(defun ebib-write-database ()
  "Writes the database to a different file."
  (interactive)
    (let ((file-name (read-file-name "Write to file: ")))
      (when file-name
	(with-temp-buffer
	  (if ebib-preamble
	      (insert (format "@preamble{%s}\n\n" ebib-preamble)))
	  (ebib-format-strings)
	  ;; we could also use mapcar on ebib-keys-list here. don't know which is faster.
	  (maphash '(lambda (x y) (ebib-format-entry x)) ebib-database)
	  (write-region (point-min) (point-max) file-name nil nil nil t))
	(setq ebib-file-name file-name)
	(setq ebib-make-backup nil)
	(ebib-set-modified nil))))

(defun ebib-save-database ()
  "Saves the database."
  (interactive)
  (if (not ebib-modified)
      (message "No changes need to be saved.")
    (let ((check-exists nil)
	  ;; if there is no file name yet, the user must enter one. if it
	  ;; already exists, write-region will ask for confirmation. if the
	  ;; user then says no, we do not want the file name he entered to be
	  ;; stored in ebib-file-name. therefore, we use in file-name.
	  (file-name ebib-file-name))
      (unless file-name
	(setq file-name (read-file-name "Save to file: "))
	(if (equal file-name "")
	    (setq file-name nil)
	  (setq check-exists t)
	  (setq ebib-make-backup nil)))
      (when file-name
	(when (and ebib-make-backup
		   (file-exists-p file-name))
	      (rename-file file-name (concat file-name "~") t)
	      (setq ebib-make-backup nil))
	(with-temp-buffer
	  (if ebib-preamble
	      (insert (format "@preamble{%s}\n\n" ebib-preamble)))
	  (ebib-format-strings)
	  (maphash '(lambda (x y) (ebib-format-entry x)) ebib-database)
	  (write-region (point-min) (point-max) file-name nil nil nil check-exists))
	;; if write-region was cancelled by the user because he didn't want
	;; to overwrite an already existing file with his new database, it
	;; throws an error, so the next lines will not be executed. hence
	;; we can now safely set ebib-file-name to the value of file-name.
	(setq ebib-file-name file-name)
	(ebib-set-modified nil)))))

(defun ebib-print-file-name ()
  "Displays the filename of the current database in the minibuffer."
  (interactive)
  (message ebib-file-name))

(defun ebib-delete-entry ()
  "Deletes the current entry from the database."
  (interactive)
  (when (and ebib-current-entry
	     (y-or-n-p (format "Delete %s? " (car ebib-current-entry))))
    (remhash (car ebib-current-entry) ebib-database)
    (with-buffer-writable
     (let ((beg (progn
		 (goto-char (overlay-start ebib-keys-overlay))
		 (point))))
       (forward-line 1)
       (delete-region beg (point))))
    (setq ebib-number-of-entries (1- ebib-number-of-entries))
    (let ((new-cur-entry (cadr ebib-current-entry)))
      (setq ebib-keys-list (delete (car ebib-current-entry) ebib-keys-list))
      (setq ebib-current-entry (member new-cur-entry ebib-keys-list)))
    (unless ebib-current-entry ; if ebib-current-entry is nil, we deleted the last entry.
      (setq ebib-current-entry (last ebib-keys-list))
      (goto-line ebib-number-of-entries))
    (ebib-set-keys-overlay)
    (ebib-fill-fields-buffer)
    (ebib-set-modified t)
    (message "Entry deleted.")))

(defun ebib-select-entry ()
  "Makes the entry at (point) the current entry."
  (interactive)
  (beginning-of-line)
    (let ((beg (point))
	  key
	  new-cur-entry)
      (save-excursion
	(end-of-line)
	(setq key (buffer-substring-no-properties beg (point))))
      (setq new-cur-entry (member key ebib-keys-list))
      (when new-cur-entry
	(setq ebib-current-entry new-cur-entry)
	(ebib-set-keys-overlay)
	(ebib-fill-fields-buffer))))

(defun ebib-export-entry ()
  "Appends the current entry to a file."
  (interactive)
  (let ((insert-default-directory (not ebib-export-file-name)))
    (setq ebib-export-file-name
	  (read-file-name
	   (format "Export %s to file: " (car ebib-current-entry))
	   "~/" nil nil ebib-export-file-name)))
  (if (equal ebib-export-file-name "")
      (setq ebib-export-file-name nil))
  (when ebib-export-file-name
    (with-temp-buffer
      (insert (format "\n")) ; to keep things tidy.
      (ebib-format-entry (car ebib-current-entry))
      (append-to-file (point-min) (point-max) ebib-export-file-name))))

(defun ebib-search (search-str)
  "Search the ebib database.

The search is conducted with STRING-MATCH and can therefore be a regexp.
Searching starts with the current entry."
  (interactive "sSearch database for: ")
  ;; maphash is not very useful here, because we want to search the
  ;; database in alphabetical order, and because we want to stop whenever a
  ;; match has been found. so we must use the keys list.
  ;; note: searching starts at the current entry. we could make it start at
  ;; the first entry, perhaps that would be more logical.
  (when search-str
    (setq ebib-search-string search-str)
    ;; first we search the current entry
    (if (ebib-search-in-entry ebib-search-string (gethash (car ebib-current-entry) ebib-database))
	(ebib-fill-fields-buffer ebib-search-string)
      ;; if the search string wasn't found in the current entry, we continue searching.
      (ebib-search-next))))

(defun ebib-search-next ()
  "Search the next occurrence of EBIB-SEARCH-STRING.
Searching starts at the entry following the current entry. If a match is
found, the matching entry is shown and becomes the new current entry."
  (interactive)
  (if (null ebib-search-string)
      (message "No search string")
    (let ((cur-search-entry (cdr ebib-current-entry)))
      (while (and cur-search-entry
		  (null (ebib-search-in-entry ebib-search-string
					      (gethash (car cur-search-entry) ebib-database))))
	(setq cur-search-entry (cdr cur-search-entry)))
      (if (null cur-search-entry)
	  (message (format "`%s' not found" ebib-search-string))
	(setq ebib-current-entry cur-search-entry)
	(set-buffer ebib-keys-buffer)
	(goto-char (point-min))
	(re-search-forward (format "^%s$" (car ebib-current-entry)))
	(beginning-of-line)
	(ebib-set-keys-overlay)
	(ebib-fill-fields-buffer ebib-search-string)))))

(defun ebib-search-in-entry (search-str entry)
  "Searches one entry of the ebib database.

Returns a list of keys in the current entry that contain the search string,
or NIL if no matches were found."
  (let ((case-fold-search t) ; we want to ensure a case-insensitive search
	(result nil))
    (maphash '(lambda (key value)
		(if (and (stringp value) ; the type* key has a symbol as value
			 (string-match search-str value))
		    (setq result (cons key result))))
	     entry)
    result))

(defun ebib-edit-strings ()
  "Edits the @STRING definitions in the database."
  (interactive)
  (ebib-fill-strings-buffer)
  (setq ebib-editing 'strings)
  (other-window 1)
  (switch-to-buffer ebib-strings-buffer)
  (goto-char (point-min)))

(defun ebib-edit-preamble ()
  "Edits the @PREAMBLE definition in the database."
  (interactive)
  (setq ebib-editing 'preamble)
  (other-window 1) ; we want the multiline edit buffer to appear in the lower window
  (ebib-multiline-edit ebib-preamble))

(defun ebib-export-preamble ()
  "Appends the @PREAMBLE defitition to a file."
  (interactive)
  (if (null ebib-preamble)
      (error "Cannot export @PREAMBLE. None defined ")
    (let ((insert-default-directory (not ebib-export-file-name)))
      (setq ebib-export-file-name
	    (read-file-name
	     "Export @PREAMBLE to file: "
	     "~/" nil nil ebib-export-file-name)))
    (if (equal ebib-export-file-name "")
	(setq ebib-export-file-name nil))
    (when ebib-export-file-name
      (with-temp-buffer
	(insert (format "\n"))		; to keep things tidy.
	(if ebib-preamble
	    (insert (format "@preamble{%s}\n\n" ebib-preamble)))
	(append-to-file (point-min) (point-max) ebib-export-file-name)))))

(defun ebib-keys-help ()
  "Displays the help message for the keys buffer."
  (interactive)
  (other-window 1)
  (ebib-display-help ebib-keys-buffer))

;;;;;;;;;;;;;;;;;
;; fields-mode ;;
;;;;;;;;;;;;;;;;;

(defvar ebib-fields-mode-map
  (make-keymap) "Keymap for the ebib fields buffer.")
(suppress-keymap ebib-fields-mode-map)

(define-key ebib-fields-mode-map [up] 'ebib-prev-field)
(define-key ebib-fields-mode-map [down] 'ebib-next-field)
(define-key ebib-fields-mode-map "k" 'ebib-prev-field)
(define-key ebib-fields-mode-map "j" 'ebib-next-field)
(define-key ebib-fields-mode-map "e" 'ebib-edit-field)
(define-key ebib-fields-mode-map "g" 'ebib-goto-first-field)
(define-key ebib-fields-mode-map "G" 'ebib-goto-last-field)
(define-key ebib-fields-mode-map [prior] 'ebib-goto-prev-set)
(define-key ebib-fields-mode-map [next] 'ebib-goto-next-set)
(define-key ebib-fields-mode-map " " 'ebib-goto-next-set)
(define-key ebib-fields-mode-map "b" 'ebib-goto-prev-set)
(define-key ebib-fields-mode-map [home] 'ebib-goto-first-field)
(define-key ebib-fields-mode-map [end] 'ebib-goto-last-field)
(define-key ebib-fields-mode-map "q" 'ebib-quit-entry-buffer)
(define-key ebib-fields-mode-map "c" 'ebib-copy-field-contents)
(define-key ebib-fields-mode-map "x" 'ebib-cut-field-contents)
(define-key ebib-fields-mode-map "p" 'ebib-paste-field-contents)
(define-key ebib-fields-mode-map "d" 'ebib-delete-field-contents)
(define-key ebib-fields-mode-map "r" 'ebib-toggle-raw)
(define-key ebib-fields-mode-map "l" 'ebib-edit-multiline-field)
(define-key ebib-fields-mode-map "s" 'ebib-insert-abbreviation)
(define-key ebib-fields-mode-map "h" 'ebib-fields-help)

(define-derived-mode ebib-fields-mode
  fundamental-mode "Ebib-fields"
  "Major mode for the Ebib fields buffer."
  (setq buffer-read-only t)
  ;; same thing with the keys
  (local-set-key "\C-xb" 'ebib-t)
  (local-set-key "\C-xk" 'ebib-t)
  (setq truncate-lines t))

(defun ebib-quit-entry-buffer ()
  "Quit editing the entry."
  (interactive)
  (other-window 1))

(defun ebib-prev-field ()
  "Move to the previous field."
  (interactive)
  (if (eq ebib-current-field 'type*)  ; if we're on the first field, just beep
      (beep)
    ;; go to the beginnig of the overlay and move upward one line.
    (goto-char (overlay-start ebib-fields-overlay))
    (forward-line -1)
    (setq ebib-current-field (prev-elem ebib-current-field ebib-cur-entry-fields))
    ;; if we ended up on an empty line, move up another line.
    (while (eq (char-after) ?\n)
	(forward-line -1))
    (ebib-set-fields-overlay)))

(defun ebib-next-field ()
  "Move to the next field."
  (interactive)
  (if (equal ebib-current-field (car (last ebib-cur-entry-fields)))
      (when (interactive-p) (beep))  ; i call this function after editing a field, and we don't want a beep then
    (goto-char (overlay-start ebib-fields-overlay))
    (forward-line 1)
    (setq ebib-current-field (next-elem ebib-current-field ebib-cur-entry-fields))
    (while (eq (char-after) ?\n)
      (forward-line 1))
    (ebib-set-fields-overlay)))

(defun ebib-goto-first-field ()
  "Move to the first field."
  (interactive)
  (setq ebib-current-field 'type*)
  (goto-char (point-min))
  (ebib-set-fields-overlay))

(defun ebib-goto-last-field ()
  "Move to the last field."
  (interactive)
  (setq ebib-current-field (car (last ebib-cur-entry-fields)))
  (goto-char (point-max))
  (forward-line -1)
  (ebib-set-fields-overlay))

(defun ebib-goto-next-set ()
  "Move to the next set of fields."
  (interactive)
  (cond
   ((eq ebib-current-field 'type*) (ebib-next-field))
   ((member ebib-current-field ebib-ign-fields) (ebib-goto-last-field))
   (t (let* ((entry-type (gethash 'type* ebib-cur-entry-hash))
	     (obl-fields (ebib-get-obl-fields entry-type))
	     (opt-fields (ebib-get-opt-fields entry-type))
	     (new-field nil))
	(when (member ebib-current-field obl-fields)
	  (setq new-field (car opt-fields)))
	(when (or (member ebib-current-field opt-fields)
		  (null new-field)) ; if ebib-current-field is among the obl-fields and there are no opt-fields
	  (setq new-field (car ebib-ign-fields)))
	(if (null new-field)
	    (ebib-goto-last-field)  ; if there was no further set to go to, go to the last field of the current set
	  (setq ebib-current-field new-field)
	  (re-search-forward (concat "^" (symbol-name ebib-current-field)))
	  (ebib-set-fields-overlay))))))

(defun ebib-goto-prev-set ()
  "Move to the previous set of fields."
  (interactive)
  (unless (eq ebib-current-field 'type*)
    (let* ((entry-type (gethash 'type* ebib-cur-entry-hash))
	   (obl-fields (ebib-get-obl-fields entry-type))
	   (opt-fields (ebib-get-opt-fields entry-type))
	   (new-field nil))
      (if (member ebib-current-field obl-fields)
	  (ebib-goto-first-field)
	(when (member ebib-current-field ebib-ign-fields)
	  (setq new-field (car (last opt-fields))))
	(when (or (member ebib-current-field opt-fields)
		  (null new-field))
	  (setq new-field (car (last obl-fields))))
	(if (null new-field)
	    (ebib-goto-first-field)
	  (setq ebib-current-field new-field)
	  (re-search-backward (concat "^" (symbol-name ebib-current-field)))
	  (ebib-set-fields-overlay))))))

(defun ebib-edit-entry-type ()
  "Edits the type of an entry."
  ;; we want to put the completion buffer in the lower window. for this
  ;; reason, we need to switch to the other window before calling
  ;; completing-read. but in order to make sure that we return to the
  ;; fields buffer and not the keys buffer when the user presses C-g, we
  ;; need to do this in an unwind-protect.
  (unwind-protect
      (progn
	(other-window 1)
	(let* ((collection (ebib-create-collection ebib-entry-types-hash))
	       (new-type (completing-read "type: " collection nil t)))
	  (unless (or (equal new-type "")
		      (not new-type))
	    (puthash 'type* (intern-soft new-type) ebib-cur-entry-hash)
	    (ebib-fill-fields-buffer)
	    (ebib-set-modified t))))
    (other-window 1)))

(defun ebib-edit-crossref ()
  "Edits the crossref field."
  (unwind-protect
      (progn
	(other-window 1)
	(let* ((collection (ebib-create-collection ebib-database))
	       (key (completing-read "Key to insert in `crossref': " collection nil t)))
	  (when key
	    (puthash 'crossref (from-raw key) ebib-cur-entry-hash)
	    (ebib-set-modified t))))
    (other-window 1)
    (ebib-redisplay-current-field)))

(defun ebib-edit-field ()
  "Edits a field of a BibTeX entry."
  (interactive)
  (cond
   ((eq ebib-current-field 'type*) (ebib-edit-entry-type))
   ((eq ebib-current-field 'crossref) (ebib-edit-crossref))
   (t
    (let ((init-contents (gethash ebib-current-field ebib-cur-entry-hash))
	  raw)
      (if (multiline-p init-contents)
	  (ebib-edit-multiline-field)
	(when init-contents
	  (if (raw-p init-contents)
	      (setq raw t)
	    (setq raw nil)
	    (setq init-contents (to-raw init-contents))))
	(let ((new-contents (read-string (format "%s: " (symbol-name ebib-current-field))
					 (if init-contents
					     (cons init-contents 0)
					   nil)
					 ebib-minibuf-hist)))
	  (when new-contents
	    (if (equal new-contents "")
		(remhash ebib-current-field ebib-cur-entry-hash)
	      (puthash ebib-current-field (if raw
					      new-contents
					    (from-raw new-contents))
		       ebib-cur-entry-hash))
	    (ebib-redisplay-current-field)
	    ;; we move to the next field, but only if ebib-edit-field was
	    ;; called interactively, otherwise we get a strange bug in
	    ;; ebib-toggle-raw...
	    (if (interactive-p) (ebib-next-field))
	    (ebib-set-modified t))))))))

(defun ebib-copy-field-contents ()
  "Copies the contents of the current field to the kill ring."
  (interactive)
  (unless (eq ebib-current-field 'type*)
    (let ((contents (gethash ebib-current-field ebib-cur-entry-hash)))
      (when (stringp contents)
	(kill-new contents)
	(message "Field contents copied.")))))

(defun ebib-cut-field-contents ()
  "Kills the contents of the current field. The killed text is put in the kill ring."
  (interactive)
  (unless (eq ebib-current-field 'type*)
    (let ((contents (gethash ebib-current-field ebib-cur-entry-hash)))
      (when (stringp contents)
	(remhash ebib-current-field ebib-cur-entry-hash)
	(kill-new contents)
	(ebib-redisplay-current-field)
	(ebib-set-modified t)
	(message "Field contents killed.")))))

(defun ebib-paste-field-contents ()
  "Inserts the last killed text into the current field.

If the current field already has a contents, nothing is inserted."
  (interactive)
  (if (eq ebib-current-field 'type*)
      (beep)
    (if (gethash ebib-current-field ebib-cur-entry-hash)
	(beep)
      (let ((new-contents (current-kill 0)))
	(when new-contents
	  (puthash ebib-current-field new-contents ebib-cur-entry-hash)
	  (ebib-redisplay-current-field)
	  (ebib-set-modified t))))))

(defun ebib-delete-field-contents ()
  "Deletes the contents of the current field. The deleted text is not put
in the kill ring."
  (interactive)
  (if (eq ebib-current-field 'type*)
      (beep)
    (remhash ebib-current-field ebib-cur-entry-hash)
    (ebib-redisplay-current-field)
    (ebib-set-modified t)
    (message "Field contents deleted.")))

(defun ebib-toggle-raw ()
  "Toggles the raw status of the current field contents."
  (interactive)
  (unless (or (eq ebib-current-field 'type*)
	      (eq ebib-current-field 'crossref))
    (let ((contents (gethash ebib-current-field ebib-cur-entry-hash)))
      (if (not contents)     ; if there is no value,
	  (progn
	    (ebib-edit-field)  ; the user can enter one, which we must then make raw
	    (let ((new-contents (gethash ebib-current-field ebib-cur-entry-hash)))
	      (when new-contents
		;; note: we don't have to check for empty string, since that is
		;; already done in ebib-edit-field
		(puthash ebib-current-field (to-raw new-contents) ebib-cur-entry-hash))))
	(if (raw-p contents)
	    (puthash ebib-current-field (from-raw contents) ebib-cur-entry-hash)
	  (puthash ebib-current-field (to-raw contents) ebib-cur-entry-hash)))
      (ebib-redisplay-current-field)
      (ebib-set-modified t))))

(defun ebib-edit-multiline-field ()
  "Edits the current field in multiline-mode."
  (interactive)
  (unless (or (eq ebib-current-field 'type*)
	      (eq ebib-current-field 'crossref))
    (let ((text (gethash ebib-current-field ebib-cur-entry-hash)))
      (if (raw-p text)
	  (setq ebib-multiline-raw t)
	(setq text (to-raw text))
	(setq ebib-multiline-raw nil))
      (ebib-multiline-edit text))))

(defun ebib-insert-abbreviation ()
  "Insert an abbreviation from the ones defined in the database."
  (interactive)
  (if (gethash ebib-current-field ebib-cur-entry-hash)
      (beep)
    (when ebib-strings-list
      (unwind-protect
	  (progn
	    (other-window 1)
	    (let* ((collection (ebib-create-collection ebib-strings))
		   (string (completing-read "Abbreviation to insert: " collection nil t)))
	      (when string
		(puthash ebib-current-field string ebib-cur-entry-hash)
		(ebib-set-modified t))))
	(other-window 1)
	;; we can't do this earlier, because we would be writing to the keys buffer...
	(ebib-redisplay-current-field)
	(ebib-next-field)))))

(defun ebib-fields-help ()
  "Displays the help message for the fields buffer."
  (interactive)
  (ebib-display-help ebib-fields-buffer))

;;;;;;;;;;;;;;;;;;
;; strings-mode ;;
;;;;;;;;;;;;;;;;;;

(defvar ebib-strings-mode-map
  (make-keymap) "Keymap for the ebib strings buffer.")
(suppress-keymap ebib-strings-mode-map)

(define-key ebib-strings-mode-map [up] 'ebib-prev-string)
(define-key ebib-strings-mode-map [down] 'ebib-next-string)
(define-key ebib-strings-mode-map "k" 'ebib-prev-string)
(define-key ebib-strings-mode-map "j" 'ebib-next-string)
(define-key ebib-strings-mode-map "e" 'ebib-edit-string)
(define-key ebib-strings-mode-map "g" 'ebib-goto-first-string)
(define-key ebib-strings-mode-map "G" 'ebib-goto-last-string)
(define-key ebib-strings-mode-map [prior] 'ebib-strings-page-up)
(define-key ebib-strings-mode-map [next] 'ebib-strings-page-down)
(define-key ebib-strings-mode-map "b" 'ebib-strings-page-up)
(define-key ebib-strings-mode-map " " 'ebib-strings-page-down)
(define-key ebib-strings-mode-map [home] 'ebib-goto-first-string)
(define-key ebib-strings-mode-map [end] 'ebib-goto-last-string)
(define-key ebib-strings-mode-map "q" 'ebib-quit-strings-buffer)
(define-key ebib-strings-mode-map "c" 'ebib-copy-string-contents)
(define-key ebib-strings-mode-map "d" 'ebib-delete-string)
(define-key ebib-strings-mode-map "a" 'ebib-add-string)
(define-key ebib-strings-mode-map "l" 'ebib-edit-multiline-string)
(define-key ebib-strings-mode-map "x" 'ebib-export-string)
(define-key ebib-strings-mode-map "X" 'ebib-export-all-strings)
(define-key ebib-strings-mode-map "h" 'ebib-strings-help)

(define-derived-mode ebib-strings-mode
  fundamental-mode "Ebib-strings"
  "Major mode for the Ebib strings buffer."
  (setq buffer-read-only t)
  ;; same thing with the keys
  (local-set-key "\C-xb" 'ebib-t)
  (local-set-key "\C-xk" 'ebib-t)
  (setq truncate-lines t))

(defun ebib-quit-strings-buffer ()
  "Quit editing the @STRING definitions."
  (interactive)
  (switch-to-buffer ebib-fields-buffer)
  (other-window 1))

(defun ebib-prev-string ()
  "Move to the previous string."
  (interactive)
  (if (equal ebib-current-string (car ebib-strings-list))  ; if we're on the first string
      (beep)
    ;; go to the beginnig of the overlay and move upward one line.
    (goto-char (overlay-start ebib-strings-overlay))
    (forward-line -1)
    (setq ebib-current-string (prev-elem ebib-current-string ebib-strings-list))
    (ebib-set-strings-overlay)))

(defun ebib-next-string ()
  "Move to the next string."
  (interactive)
  (if (equal ebib-current-string (car (last ebib-strings-list)))
      (when (interactive-p) (beep))
    (goto-char (overlay-start ebib-strings-overlay))
    (forward-line 1)
    (setq ebib-current-string (next-elem ebib-current-string ebib-strings-list))
    (ebib-set-strings-overlay)))

(defun ebib-goto-first-string ()
  "Move to the first string."
  (interactive)
  (setq ebib-current-string (car ebib-strings-list))
  (goto-char (point-min))
  (ebib-set-strings-overlay))

(defun ebib-goto-last-string ()
  "Move to the last string."
  (interactive)
  (setq ebib-current-string (car (last ebib-strings-list)))
  (goto-char (point-max))
  (forward-line -1)
  (ebib-set-strings-overlay))

(defun ebib-strings-page-up ()
  "Moves 10 entries up in the database."
  (interactive)
  (let ((number-of-strings (length ebib-strings-list))
	(remaining-number-of-strings (length (member ebib-current-string ebib-strings-list))))
    (if (<= (- number-of-strings remaining-number-of-strings) 10)
	(ebib-goto-first-string)
      (setq ebib-current-string (nth
				 (- number-of-strings remaining-number-of-strings 10)
				 ebib-strings-list))
      (goto-char (overlay-start ebib-strings-overlay))
      (forward-line -10)
      (ebib-set-strings-overlay)))
  (message ebib-current-string))

(defun ebib-strings-page-down ()
  "Moves 10 entries down in the database."
  (interactive)
  (let ((number-of-strings (length ebib-strings-list))
	(remaining-number-of-strings (length (member ebib-current-string ebib-strings-list))))
    (if (<= remaining-number-of-strings 10)
      (ebib-goto-last-string)
    (setq ebib-current-string (nth
			      (- number-of-strings remaining-number-of-strings -10)
			      ebib-strings-list))
    (goto-char (overlay-start ebib-strings-overlay))
    (forward-line 10)
    (ebib-set-strings-overlay)))
  (message ebib-current-string))


(defun ebib-fill-strings-buffer ()
  "Fills the strings buffer with the @string definitions."
  (set-buffer ebib-strings-buffer)
  (with-buffer-writable
   (erase-buffer)
   (dolist (elem ebib-strings-list)
     (let ((str (to-raw (gethash elem ebib-strings))))
       (insert (format "%-18s %s\n" elem
		       (if (multiline-p str)
			   (concat "+" (first-line str))
			 (concat " " str)))))))
  (goto-char (point-min))
  (setq ebib-current-string (car ebib-strings-list))
  (ebib-set-strings-overlay)
  (set-buffer-modified-p nil))

(defun ebib-edit-string ()
  "Edits the value of an @STRING definition

When the user enters an empty string, the value is not changed."
  (interactive)
  (let ((init-contents (to-raw (gethash ebib-current-string ebib-strings))))
    (if (multiline-p init-contents)
	(ebib-edit-multiline-string)
      (let ((new-contents (read-string (format "%s: " ebib-current-string)
				       (if init-contents
					   (cons init-contents 0)
					 nil)
				       ebib-minibuf-hist)))
	(when new-contents
	  (if (equal new-contents "")
	      (error "@STRING definition cannot be empty")
	    (puthash ebib-current-string (from-raw new-contents) ebib-strings)
	    (ebib-redisplay-current-string)
	    (ebib-next-string)
	    (ebib-set-modified t)))))))

(defun ebib-copy-string-contents ()
  "Copies the contents of the current string to the kill ring."
  (interactive)
  (let ((contents (gethash ebib-current-string ebib-strings)))
    (kill-new contents)
    (message "String value copied.")))

(defun ebib-delete-string ()
  "Deletes the current @STRING definition from the database."
  (interactive)
  (when (y-or-n-p (format "Delete @STRING definition %s? " ebib-current-string))
    (remhash ebib-current-string ebib-strings)
    (with-buffer-writable
     (let ((beg (progn
		 (goto-char (overlay-start ebib-strings-overlay))
		 (point))))
     (forward-line 1)
     (delete-region beg (point))))
    (let ((new-cur-string (next-elem ebib-current-string ebib-strings-list)))
      (setq ebib-strings-list (delete ebib-current-string ebib-strings-list))
      (when (null new-cur-string)	; deleted the last string
	(setq new-cur-string (car (last ebib-strings-list)))
	(forward-line -1))
      (setq ebib-current-string new-cur-string))
    (ebib-set-strings-overlay)
    (ebib-set-modified t)
    (message "@STRING definition deleted.")))

(defun ebib-add-string ()
  "Creates a new @STRING definition."
  (interactive)
  (let ((new-abbr (read-string "New @STRING abbreviation: ")))
    (if (equal new-abbr "")
	(setq new-abbr nil))
    (if (member new-abbr ebib-strings-list)
	(error (format "%s already exists" new-abbr))
      (when new-abbr
	(let ((new-string (read-string (format "Value for %s: " new-abbr))))
	  (if (equal new-string "")
	      (setq new-string nil))
	  (when new-string
	    (puthash new-abbr (from-raw new-string) ebib-strings)
	    (setq ebib-strings-list (sort (cons new-abbr ebib-strings-list) 'string<))
	    (sort-in-buffer (length ebib-strings-list) new-abbr)
	    (with-buffer-writable
	      (insert (format "%-19s %s\n" new-abbr new-string)))
	    (forward-line -1)
	    (ebib-set-strings-overlay)
	    (setq ebib-current-string new-abbr)
	    (ebib-set-modified t)))))))

(defun ebib-export-string ()
  "Appends the current @STRING definition to a file."
  (interactive)
  (when ebib-current-string
    (let ((insert-default-directory (not ebib-export-file-name)))
      (setq ebib-export-file-name
	    (read-file-name
	     (format "Export @STRING definition `%s' to file: " ebib-current-string)
	     "~/" nil nil ebib-export-file-name)))
    (if (equal ebib-export-file-name "")
	(setq ebib-export-file-name nil))
    (when ebib-export-file-name
      (with-temp-buffer
	(insert (format "\n@string{%s = %s}\n" ebib-current-string (gethash ebib-current-string ebib-strings)))
	(append-to-file (point-min) (point-max) ebib-export-file-name)))))

(defun ebib-export-all-strings ()
  "Appends all the @STRING definitions to a file."
  (interactive)
  (when ebib-current-string  ; there is always a current string, unless there are no strings
    (let ((insert-default-directory (not ebib-export-file-name)))
      (setq ebib-export-file-name
	    (read-file-name
	     "Export all @STRING definitions to file: "
	     "~/" nil nil ebib-export-file-name)))
    (if (equal ebib-export-file-name "")
	(setq ebib-export-file-name nil))
    (when ebib-export-file-name
      (with-temp-buffer
	(insert (format "\n"))		; to keep things tidy.
	(ebib-format-strings)
	(append-to-file (point-min) (point-max) ebib-export-file-name)))))

(defun ebib-edit-multiline-string ()
  "Edits the current string in multiline-mode."
  (interactive)
  (ebib-multiline-edit (to-raw (gethash ebib-current-string ebib-strings))))

(defun ebib-strings-help ()
  "Displays the help message for the strings buffer."
  (interactive)
  (ebib-display-help ebib-strings-buffer))

;;;;;;;;;;;;;;;;;;;;;;;;;
;; multiline-edit-mode ;;
;;;;;;;;;;;;;;;;;;;;;;;;;

(define-derived-mode ebib-multiline-edit-mode
  text-mode "Ebib-edit"
  "Major mode for editing multiline strings in Ebib."
  ;; we redefine some basic keys because we need them to leave this buffer.
  ;; again with the cheating!
  (local-set-key "\C-xb" 'ebib-store-multiline-text)
  (local-set-key "\C-x\C-s" 'ebib-store-multiline-text)
  (local-set-key "\C-xk" 'ebib-cancel-multiline-edit))

(defun ebib-multiline-edit (&optional starttext)
  "Switches to Ebib's multiline edit buffer.

STARTTEXT is a string that contains the initial text of the buffer."
  ;; note: the buffer is put in the currently active window!
  (switch-to-buffer ebib-multiline-buffer)
  (erase-buffer)
  (when starttext
    (insert starttext)
    (goto-char (point-min))
    (set-buffer-modified-p nil)))

(defun ebib-store-multiline-text ()
  "Quits the multiline edit buffer and stores the text."
  (interactive)
  (when (y-or-n-p "Store the text? ")
    (let ((text (buffer-substring-no-properties (point-min) (point-max))))
      (cond
       ((eq ebib-editing 'preamble)
	(if (equal text "")
	    (setq ebib-preamble nil)
	  (setq ebib-preamble text))
	(switch-to-buffer ebib-fields-buffer)
	(other-window 1)) ; have to switch back to the keys buffer window
       ((eq ebib-editing 'fields)
	(if (equal text "")
	    (remhash ebib-current-field ebib-cur-entry-hash)
	  (when (not ebib-multiline-raw)
	    (setq text (from-raw text)))
	  (puthash ebib-current-field text ebib-cur-entry-hash))
	(switch-to-buffer ebib-fields-buffer)
	(ebib-redisplay-current-field)
	(ebib-next-field))
       ((eq ebib-editing 'strings)
	(if (equal text "")
	    ;; with ERROR, we avoid execution of EBIB-SET-MODIFIED and
	    ;; MESSAGE, but we also do not switch back to the strings
	    ;; buffer. this may not be so bad, actually, because the user
	    ;; may want to change his edit.
	    (error "@STRING definition cannot be empty ")
	  (setq text (from-raw text))  ; strings cannot be raw
	  (puthash ebib-current-string text ebib-strings))
	(switch-to-buffer ebib-strings-buffer)
	(ebib-redisplay-current-string)
	(ebib-next-string))))
    (ebib-set-modified t)
    (message "Text stored")))

(defun ebib-cancel-multiline-edit ()
  "Quits the multiline edit buffer and discards the changes."
  (interactive)
  (when (y-or-n-p "Cancel edit? ")
    (cond
     ((eq ebib-editing 'fields)
      (switch-to-buffer ebib-fields-buffer))
     ((eq ebib-editing 'strings)
      (switch-to-buffer ebib-strings-buffer))
     ((eq ebib-editing 'preamble)
      (switch-to-buffer ebib-fields-buffer)
      (other-window 1)))
    (message "Multiline edit cancelled")))

;;;;;;;;;;;;;;;;;;;;
;; ebib-help-mode ;;
;;;;;;;;;;;;;;;;;;;;

(defvar ebib-help-mode-map
  (make-keymap) "Keymap for the ebib help buffer.")
(suppress-keymap ebib-help-mode-map)

(define-key ebib-help-mode-map " " 'scroll-up)
(define-key ebib-help-mode-map "b" 'scroll-down)
(define-key ebib-help-mode-map "q" 'ebib-quit-help-buffer)

(define-derived-mode ebib-help-mode
  fundamental-mode "Ebib-help"
  "Major mode for the Ebib help buffer."
  (setq buffer-read-only t)
  ;; same thing with the keys
  (local-set-key "\C-xb" 'ebib-quit-help-buffer)
  (local-set-key "\C-xk" 'ebib-quit-help-buffer))

(defun ebib-display-help (buffer)
  "Shows the help message for Ebib-buffer BUFFER."
  (switch-to-buffer ebib-help-buffer)
  (setq ebib-help buffer)
  (with-buffer-writable
   (erase-buffer)
   (cond
    ((eq buffer ebib-keys-buffer) (insert ebib-keys-buffer-help))
    ((eq buffer ebib-fields-buffer) (insert ebib-fields-buffer-help))
    ((eq buffer ebib-strings-buffer) (insert ebib-strings-buffer-help)))
   (goto-char (point-min))))

(defun ebib-quit-help-buffer ()
  "Exitst the help buffer."
  (interactive)
  (cond
   ((eq ebib-help ebib-keys-buffer)
    (switch-to-buffer ebib-fields-buffer)
    (other-window 1))
   ((eq ebib-help ebib-fields-buffer)
    (switch-to-buffer ebib-fields-buffer))
   ((eq ebib-help ebib-strings-buffer)
    (switch-to-buffer ebib-strings-buffer))))

;;;;;;;;;;;;;;;;;;;;;;;;;
;; one global function ;;
;;;;;;;;;;;;;;;;;;;;;;;;;

(defun ebib-insert-bibtex-key ()
  "Inserts a BibTeX key at POINT, surrounded by braces.

The user is prompted for a BibTeX key and has to choose one from the
database loaded in Ebib. Tab completion works."
  (interactive)
  (if (or
       (null ebib-database)
       (= (hash-table-count ebib-database) 0))
      (error "No entries in database")
    (let* ((collection (ebib-create-collection ebib-database))
	   (key (completing-read "Key to insert: " collection nil t nil ebib-minibuf-hist)))
      (when key
	(insert (format "{%s}" key))))))

(provide 'ebib)

;; we put these at the end, because they seem to mess up Emacs'
;; font-highlighting.

(defconst ebib-keys-buffer-help 
  "Ebib keys buffer -- command key overview

Note: command keys are case-sensitive.

(Press C-v to scroll down, M-v to scroll up, `q' to quit.)

cursor movement:
[up], k:             go to the previous entry
[down], j:           go to the next entry
[home], g:           go to the first entry
[end], G:            go to the last entry
[PgUp], b:           scroll 10 entries up
[PgDn], [space]:     scroll  10 entries down

editing:
e:                   edit the current entry
a:                   add a new entry
d:                   delete the current entry
S:                   edit the @STRING definitions
P:                   edit the @PREAMBLE definition

searching:
/:                   search the database
n:                   find the next occurrence of the search string
C-s:                 search for a key (incrementally)
[return]:            select the entry under the cursor (use after C-s)

file handling:
o:                   open a database
c:                   close the database
s:                   save the database
w:                   save the database under a different name
m:                   merge another database
x:                   export the current entry to another file
X:                   export the @PREAMBLE definition to another file
f:                   print the full file name in the minibuffer

z:                   put Ebib in the background
q:                   quit Ebib
")

(defconst ebib-fields-buffer-help
  "Ebib fields buffer -- command key overview

Note: command keys are case-sensitive.

(Press C-v to scroll down, M-v to scroll up, `q' to quit.)

cursor movement:
[up], k:             go to the previous field
[down], j:           go to the next field
[home], g:           go to the first field
[last], G:           go to the last field
[PgUp], b:           go to the previous group of fields
[PgDn], [space]:     go to the next group of fields

editing:
e:                   edit the value of the current field
c:                   copy the value of the current field (value is put into the kill ring)
x:                   kill the value of the current field (value is put into the kill ring)
p:                   paste the most recently copied/cut string
d:                   delete the value of the current entry
r:                   toggle the \"rawness\" status of the current field
l:                   edit the current field as multi-line
s:                   insert an @STRING abbreviation into the current field

q:                   quit the fields buffer and return to the keys buffer
")

(defconst ebib-strings-buffer-help
  "Ebib strings buffer -- command key overview

Note: command keys are case-sensitive.

(Press C-v to scroll down, M-v to scroll up, `q' to quit.)

cursor movement:
[up], k:             go to the previous @STRING definition
[down], j:           go to the next @STRING definition
[home], g:           go to the first @STRING definition
[end], G:            go to the last @STRING definition
[PgUp], b:           scroll 10 @STRING definitions up
[PgDn], [space]:     scroll 10 @STRING definitions down

editing:
e:                   edit the value of the current @STRING definition
c:                   copy the value of the current @STRING definition
d:                   delete the current @STRING definition
a:                   add an @STRING definition
l:                   edit the current @STRING as multi-line

exporting:
x:                   export the current @STRING definition to another file
X:                   export all @STRING definitions to another file

q:                   quit the strings buffer and return to the keys buffer
")


;;; ebib ends here
